#ifndef _HLIST_H_
#define _HLIST_H_

#include "common.h"

/*
 * Double-linked list implementation
 */
struct hlist_elem;

/**
 * Double-linked list structure
 */
struct hlist_head {
	struct hlist_elem *first;	/*!< pointer to list head */
};

struct hlist_elem {
	struct hlist_elem *next;	/*!< next element (first) */
	struct hlist_elem *prev;	/*!< previous element (last) */
};

/**
 * hlist_init_head
 * \brief	Initialize linked list head
 * \param[in]	h	pointer to linked list head
 */
void hlist_init_head( struct hlist_head *h );

/**
 * hlist_init_elem
 * \brief	Initialize linked list element
 * \param[in]	e	pointer to linked list element
 */
void hlist_init_elem( struct hlist_elem *e );

/**
 * hlist_entry
 * \brief Get current element
 */
#define hlist_entry(ptr, type, member) \
	container_of(ptr, type, member)



/**
 * hlist_insert_after
 * \brief	Insert new element after given one
 * \param[in]	elem	Element to insert after
 * \param[in]	new	New element
 */
void hlist_insert_after( struct hlist_head *liust, struct hlist_elem *elem, struct hlist_elem *enew );


/**
 * hlist_insert_head
 * \brief	Insert new element at first position
 * \param[in]	list	The list
 * \param[in]	new	New element
 */
#define hlist_insert_head(list, enew) \
	hlist_insert_after(list, (list)->first, enew)


/**
 * hlist_insert_before
 * \brief	Insert new element before given one
 * \param[in]	elem	Element to insert before
 * \param[in]	new	New element
 */
void hlist_insert_before( struct hlist_head *list, struct hlist_elem *elem, struct hlist_elem *enew );


/**
 * hlist_insert_tail
 * \brief	Insert new element at the end of the list
 * \param[in]	list	The list
 * \param[in]	new	New element
 */
void hlist_insert_tail( struct hlist_head *list, struct hlist_elem *enew );


/**
 * hlist_remove
 * \brief	Remove element from the list
 * \param[in]	elem	Element to remove
 */
void hlist_remove( struct hlist_head *list, struct hlist_elem *elem );


/**
 * hlist_first
 * \brief	Get first element in the list
 * \param[in]	elem	Any list element
 * \return	First element
 */
struct hlist_elem * hlist_first( struct hlist_elem *elem );
struct hlist_elem * hlist_first_elem( struct hlist_head *list );


/**
 * hlist_last
 * \brief	Get last element in the list
 * \param[in]	elem	Any list element
 * \return	Last element
 */
struct hlist_elem * hlist_last( struct hlist_elem *elem );
struct hlist_elem * hlist_last_elem( struct hlist_head *list );


/**
 * hlist_next
 * \brief	Get next element after given one
 * \param[in]	elem	Any list element
 * \return	Next element
 */
struct hlist_elem * hlist_next( struct hlist_elem *elem );


/**
 * hlist_prev
 * \brief	Get element before given one
 * \param[in]	elem	Any list element
 * \return	Previous element
 */
struct hlist_elem * hlist_prev( struct hlist_elem *elem );

/**
 * hlist_is_first
 * \brief	Check if the element passed is the first one
 * \param[in]	elem	Any list element
 * \return	0 if the element passed is not the first, 1 otherwise
 */
int hlist_is_first( struct hlist_elem *elem );


/**
 * hlist_is_last
 * \brief	Check if the element passed is the last one
 * \param[in]	elem	Any list element
 * \return	0 if the element passed is not the last, 1 otherwise
 */
int hlist_is_last( struct hlist_elem *elem );


/**
 * hlist_replace
 * \brief	Replace an element in the list with a new one
 * \param[in]	old	Element to replace
 * \param[in]	new	Replacement
 */
void hlist_replace( struct hlist_head *list, struct hlist_elem *eold, struct hlist_elem *enew );


/**
 * hlist_for_each
 * \brief	Walk the list of elements in a for like loop
 * \param[in]	head	The list
 * \param[in]	p	Pointer to store the current element
 * \note	Not hlist_remove() safe!
 */
#define hlist_for_each(head, p)			\
	for (p = (typeof(*p) *)(head)->first;	\
	p; p = (typeof(*p) *)(p)->next)


/**
 * hlist_for_each_safe
 * \brief	Walk the list of elements in a for like loop
 * \param[in]	head	The list
 * \param[in]	n	Temporary storage
 * \param[in]	p	Pointer to store the current element
 * \note	hlist_remove() safe version
 */
#define hlist_for_each_safe(head, n, p)	\
	for (p = (typeof(*p) *)(head)->first, n = (typeof(*p) *)(p)->next; \
	p; p = (typeof(*p) *)(n), n = (typeof(*p) *)(p)->next)


/**
 * hlist_is_empty
 * \brief	Check if the list has no elements
 * \param[in]	head	The list
 * \return	0 if the list is not empty, 1 otherwise
 */
#define hlist_is_empty(head) ({		\
	((head)->first == NULL);	\
	})


/**
 * hlist_num_entries
 * \brief	Get the number of elements in the list
 * \param[in]	head	The list
 * \return	number of entries
 */
#define hlist_num_entries(head) ({		\
	int _x = 0;				\
	struct hlist_elem *p = (head)->first;	\
	for (; p; _x++) p = p->next;	\
	_x; })


#endif

