#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdarg.h>

#include "tdm_private.h"
#include "tdm_common.h"
#include "tdm_module.h"
#include "tdm_log.h"

#define TDM_MODULES_DIR    "/home/stkn/Stuff/libtdm/src"
#define TDM_MODULES_PREFIX "mod_"
#define TDM_MODULES_EXT    "so"

#include <dirent.h>
#include <dlfcn.h>

static int load_module(const char *path)
{
	struct tdm_module *info = NULL;
	void *ptr = NULL;
	int res = -1;

	tdm_notice("Loading module '%s'\n", path);

	ptr = dlopen(path, RTLD_NOW | RTLD_LOCAL);
	if (ptr == NULL) {
		tdm_error("Failed to load module '%s': %s\n", path, dlerror());
		goto out;
	}

	if ((info = dlsym(ptr, "tdm_module")) == NULL) {
		tdm_error("Module symbol not found: %s\n", dlerror());
		goto out_error;
	}

	tdm_info("Successfully loaded module '%s'\n", path);

	if (!info->init || !info->name) {
		tdm_error("Mandatory minimum information/functions not supplied\n");
		goto out_error;
	}

	if (info->name)
		tdm_info("Module name:     %s\n", info->name);
	if (info->init)
		tdm_info("Init function:   %p\n", info->init);
	if (info->fini)
		tdm_info("Fini function:   %p\n", info->fini);
	if (info->reload)
		tdm_info("Reload function: %p\n", info->reload);
	if (info->probe)
		tdm_info("Probe function:  %p\n", info->probe);

	tdm_info("Initializing module '%s'\n", info->name);

	if ((res = info->init(NULL)) != TDM_STATUS_SUCCESS) {
		tdm_error("Failed to inizialize module '%s'\n", info->name);
		goto out_error;
	}
	res = 0;

out:
	return res;

out_error:
	dlclose(ptr);
	return res;
}

static int load_modules(void)
{
	struct dirent entry, *result;
	int count = 0;
	DIR *dir = NULL;

	dir = opendir(TDM_MODULES_DIR);
	if (!dir) {
		tdm_error("Failed to open modules directory\n");
		return -1;
	}

	while (readdir_r(dir, &entry, &result) == 0) {
		char path[PATH_MAX] = { '\0' };
		char *extptr = NULL;

		if (!result) break;

		/* only files */
		if (result->d_type != DT_REG)
			continue;

		/* no extension */
		if ((extptr = strrchr(result->d_name, '.')) == NULL)
			continue;

		/* not a module? */
		if (strncmp(result->d_name, TDM_MODULES_PREFIX, sizeof(TDM_MODULES_PREFIX) - 1) || strncmp(++extptr, TDM_MODULES_EXT, sizeof(TDM_MODULES_EXT) - 1))
			continue;

		tdm_info("Found module '%s'\n", result->d_name);

		/* module already loaded? */

		/* no, module in ignore list? */

		/* no, try to load module */
		snprintf(path, sizeof(path), "%s/%s", TDM_MODULES_DIR, result->d_name);
		load_module(path);

		/* increase counter */
		count++;
	}

	if (closedir(dir) < 0) {
		tdm_error("Failed to close modules directory handle\n");
	}

	tdm_notice("Found %d modules\n", count);

	return count;
}

/*
 *
 */
int main(void)
{
	tdm_log_init();
	tdm_set_loglevel(TDM_LOG_INFO);

	load_modules();
	return 0;
}
